﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using System.Web.Http;
using BL.DTOs.Reviews;
using BL.Facades;
using Newtonsoft.Json;

namespace API.Controllers
{
    public class RatingController : ApiController
    {
        public ProductFacade ProductFacade { get; set; }

        /// <summary>
        /// GET: api/Rating
        /// Gets all reviews (regardless of product)
        /// </summary>
        /// <returns>HTTP 200 on success</returns>
        public IEnumerable<ProductReviewDTO> Get()
        {
            return ProductFacade.GetAllReviews();
        }

        /// <summary>
        /// GET: api/Rating/Product/1
        /// Gets all reviews for corresponding product
        /// </summary>
        /// <param name="id">id of product to get reviews from</param>
        /// <returns>HTTP 200 on success</returns>
        [Route("~/api/Rating/Product/{id}")]
        public IEnumerable<ProductReviewDTO> Get(int id)
        {
            return id <= 0 ? null : ProductFacade.GetAllReviews(id);
        }

        /// <summary>
        /// POST: api/Rating
        /// Creates new review
        /// </summary>
        /// <param name="value">review data</param>
        /// <returns>HTTP 201 on success</returns>
        public IHttpActionResult Post([FromBody]string value)
        {
            try
            {
                var review = JsonConvert.DeserializeObject<ProductReviewCreateDTO>(value);
                ProductFacade.AddReview(review);
                return Content(HttpStatusCode.Created, review);
            }
            catch (JsonException)
            {
                Debug.WriteLine($"Rating API - Post(...) - failed to deserialize value: {value}");
                return StatusCode(HttpStatusCode.BadRequest);
            }
            catch (NullReferenceException ex)
            {
                Debug.WriteLine(ex.Message);
                return StatusCode(HttpStatusCode.NotFound);
            }
            
        }

        /// <summary>
        /// PUT: api/Rating/1
        /// Updates corresponding review
        /// </summary>
        /// <param name="id">id of review to update</param>
        /// <param name="value">modified review data</param>
        /// <returns>HTTP 200 on success</returns>
        public IHttpActionResult Put(int id, [FromBody]string value)
        {
            try
            {
                var productReviewEditDto = JsonConvert.DeserializeObject<ProductReviewEditDTO>(value);
                if (id <= 0)
                {
                    return Content(HttpStatusCode.PreconditionFailed, "Product review ID must be greater than zero.");
                }
                productReviewEditDto.ID = id;
                ProductFacade.EditReview(productReviewEditDto);
                return Content(HttpStatusCode.OK, productReviewEditDto);
            }
            catch (JsonException)
            {
                Debug.WriteLine($"Rating API - Put(...) - failed to deserialize value: {value}");
                return StatusCode(HttpStatusCode.BadRequest);
            }
            catch (NullReferenceException ex)
            {
                Debug.WriteLine(ex.Message);
                return StatusCode(HttpStatusCode.NotFound);
            }
           
        }

        /// <summary>
        /// DELETE: api/Rating/1
        /// Deletes corresponding review
        /// </summary>
        /// <param name="id">id of review to delete</param>
        /// <returns>HTTP 204 on success</returns>
        public IHttpActionResult Delete(int id)
        {
            try
            {
                ProductFacade.DeleteReview(id);
                return StatusCode(HttpStatusCode.NoContent);
            }
            catch (NullReferenceException ex)
            {
                Debug.WriteLine(ex.Message);
                return StatusCode(HttpStatusCode.NotFound);
            }
        }
    }
}
