﻿using System;
using System.Collections.Generic;
using BL.DTOs.Coupons;
using BL.DTOs.Filters;
using BL.DTOs.Orders;
using BL.DTOs.Stocks;
using BL.Services.Checkout;
using BL.Services.Sales;
using BL.Services.Stocks;

namespace BL.Facades
{
    public class OrderFacade
    {
        #region Dependencies

        private readonly ICheckoutService checkoutService;

        private readonly IStockService stocksService;

        private readonly ISalesService salesService;

        public OrderFacade(IStockService stocksService, ICheckoutService checkoutService, ISalesService salesService)
        {
            this.stocksService = stocksService;
            this.checkoutService = checkoutService;
            this.salesService = salesService;
        }

        #endregion

        #region OrderManagement

        /// <summary>
        /// Persists order together with all related data
        /// </summary>
        /// <param name="createOrderDto">wrapper for order, orderItems, customer and coupon</param>
        /// <param name="confirmationAction">Action to perform when uow successfully performs after commit action</param>
        public void ConfirmOrder(OrderCreateDTO createOrderDto, Action confirmationAction)
        {
            checkoutService.ConfirmOrder(createOrderDto, confirmationAction);
        }

        /// <summary>
        /// Gets all OrderItems from specified order
        /// </summary>
        /// <param name="filter">order item filter</param>
        /// <param name="requiredPage">page to show</param>
        /// <returns>all order items from specified order</returns>
        public IEnumerable<OrderItemDTO> ListItemsFromOrder(OrderItemFilter filter, int requiredPage = 1)
        {
            return salesService.ListItemsFromOrder(filter, requiredPage);
        }

        /// <summary>
        /// Calculates total price for all order items (overall coupon discount is included)
        /// </summary>
        /// <param name="orderItems">all order items</param>
        /// <param name="couponId">coupon id</param>
        /// <returns>Total price for given items</returns>
        public decimal CalculateTotalPrice(ICollection<OrderItemDTO> orderItems, int couponId = 0)
        {
            return checkoutService.CalculateTotalPrice(orderItems, couponId);
        }

        /// <summary>
        /// Gets discount coupon according to promo code
        /// </summary>
        /// <param name="promoCode">Coupon promo code</param>
        /// <param name="customerId">ID of customer who owns the coupon</param>
        /// <returns>Respective discount coupon</returns>
        public CouponDTO GetCoupon(string promoCode, int customerId)
        {
            return checkoutService.GetCoupon(promoCode, customerId);
        }

        /// <summary>
        /// Notifies warehouse about realised order and requests corresponding order shipment
        /// </summary>
        public void RequestOrderDelivery()
        {
            /*  for instance a request to https://www.easypost.com/dhl-express-api 
                can be made, corresponding warehouse receives a notification about
                delivery request and schedules delivery
            */
        } 

        #endregion

        #region StocksManagement

        /// <summary>
        /// Performs stock reservation in order to ensure
        /// simultaneous orders can be satisfied
        /// </summary>
        /// <param name="stockReservation">Stock reservation details containing 
        /// customer reserving the stock, amount, stockID and expiration</param>
        /// <returns>true if reservation is successfull, otherwise false</returns>
        public bool ReserveStock(StockReservationDTO stockReservation)
        {
            return stocksService.ReserveStock(stockReservation);
        }

        /// <summary>
        /// Performs stock release 
        /// (typically when user confirms or cancels the order)
        /// </summary>
        /// <param name="customerId">ID of customer who reserved the stock</param>
        /// <param name="stockIds">IDs of the stock to release</param>
        public void ReleaseReservations(int customerId, params int[] stockIds)
        {
            stocksService.ReleaseStockReservations(customerId, stockIds);
        }

        /// <summary>
        /// Performs stock expedition by decreasing stocks 
        /// stored units and releasing corresponding reservations
        /// </summary>
        /// <param name="customerId">Id of customer whose reserved stock should be dispatched</param>
        public void DispatchStock(int customerId)
        {
            stocksService.DispatchStock(customerId);
        }

        /// <summary>
        /// Performs order of required stock from distributor
        /// </summary>
        /// <param name="stockId">id of stock to order</param>
        /// <param name="quantity">Number of units to order</param>
        public void OrderStockFromDistributor(int stockId, int quantity)
        {
            stocksService.OrderStockFromDistributor(stockId, quantity);
        }

        /// <summary>
        /// Gets number of currently available (free)
        /// units for stock belonging to product with specified ID
        /// </summary>
        /// <param name="stockId">ID of the stock for which stock availability should be checked</param>
        /// <returns>Number of free units for this product</returns>
        public int GetCurrentlyAvailableUnits(int stockId)
        {
            return stocksService.GetCurrentlyAvailableUnits(stockId);
        }
        
        #endregion
    }
}
