﻿using System.Collections.Generic;
using System.Linq;
using BL.DTOs.Categories;
using BL.DTOs.Filters;
using BL.DTOs.Orders;
using BL.DTOs.Products;
using BL.DTOs.Reviews;
using BL.Services.Categories;
using BL.Services.Checkout;
using BL.Services.Products;
using BL.Services.Rating;
using BL.Services.Stocks;

namespace BL.Facades
{
    public class ProductFacade
    {
        public int ProductPageSize => productService.ProductsPageSize;

        #region Dependencies

        private readonly ICategoryService categoryService;

        private readonly IProductService productService;

        private readonly IStockService stocksService;

        private readonly IRatingService ratingService;

        private readonly ICheckoutService checkoutService;

        public ProductFacade(ICategoryService categoryService, IProductService productService,
            IStockService stocksService, IRatingService ratingService, ICheckoutService checkoutService)
        {
            this.categoryService = categoryService;
            this.productService = productService;
            this.stocksService = stocksService;
            this.ratingService = ratingService;
            this.checkoutService = checkoutService;
        }

        #endregion

        #region ProductManagement

        /// <summary>
        /// Get product according to ID
        /// </summary>
        /// <param name="productId">ID of the product</param>
        /// <returns>The product with given ID, null otherwise</returns>
        public ProductDTO GetProduct(int id)
        {
            var product = productService.GetProduct(id);
            product.CurrentlyAvailableUnits = stocksService.GetCurrentlyAvailableUnits(product.StockID);
            return product;
        }

        /// <summary>
        /// Gets products according to filter and required page
        /// </summary>
        /// <param name="filter">products filter</param>
        /// <param name="requiredPage">page to show</param>
        /// <returns></returns>
        public ProductListQueryResultDTO GetProducts(ProductFilter filter, int requiredPage = 1)
        {
            var productListQueryResult = productService.ListProducts(filter, requiredPage);
            foreach (var result in productListQueryResult.ResultsPage)
            {
                result.CurrentlyAvailableUnits = stocksService.GetCurrentlyAvailableUnits(result.StockID);
            }
            return productListQueryResult;
        }

        /// <summary>
        /// Gets detailed information about product with given ID
        /// </summary>
        /// <param name="id">ID of the product</param>
        /// <returns>Detailed information about product with given ID</returns>
        public ProductDetailDTO GetProductDetails(int id)
        {
            var productDto = GetProduct(id);
            var orderItem = new OrderItemDTO { Product = productDto, Quantity = 1 };
            var productFinalPrice = checkoutService.GetItemsTotalPrice(orderItem);
            var reviews = ratingService.GetAllReviewsForProduct(id).ToList();

            if (productDto != null)
            {
                productDto.AverageRating = reviews.Any() ? reviews.Average(review => review.UserRating) as double? : null;
            }

            return new ProductDetailDTO
            {
                Product = productDto,
                ProductFinalPrice = productFinalPrice,
                Reviews = reviews
            };
        }
        
        /// <summary>
        /// Creates product with category that corresponds with given name
        /// </summary>
        /// <param name="product">product</param>
        /// <param name="categoryName">category name</param>
        public void CreateProductWithCategoryName(ProductDTO product, string categoryName)
        {
            var categoryId = categoryService.GetProductCategoryIdByName(categoryName);
            productService.CreateProduct(product, categoryId);
        }

        public void EditProduct(ProductDTO productDto, int categoryId, params int[] productReviewIds)
        {
            productService.EditProduct(productDto, categoryId, productReviewIds);
        }

        public void DeleteProduct(int id)
        {
            productService.DeleteProduct(id);
        }

        #endregion

        #region CategoriesManagement

        /// <summary>
        /// Gets category according to ID
        /// </summary>
        /// <param name="categoryId">category ID</param>
        /// <returns>The category</returns>
        public CategoryDTO GetCategory(int categoryId)
        {
            return categoryService.GetCategory(categoryId);
        }

        /// <summary>
        /// Gets category id that belongs to specified category name
        /// </summary>
        /// <param name="name">category name</param>
        /// <returns>id of category with specified name</returns>
        public int GetProductCategoryIdByName(string name)
        {
            return categoryService.GetProductCategoryIdByName(name);
        }

        /// <summary>
        /// Gets all categories
        /// </summary>
        /// <returns>All available categories</returns>
        public IEnumerable<CategoryDTO> GetAllCategories()
        {
            return categoryService.ListAllCategories();
        }

        #endregion

        #region ProductReviews
        public void AddReview(ProductReviewCreateDTO reviewDto)
        {
            ratingService.AddReview(reviewDto);
        }

        public void EditReview(ProductReviewEditDTO reviewDto)
        {
            ratingService.EditReview(reviewDto);
        }

        public void DeleteReview(int reviewId)
        {
            ratingService.DeleteReview(reviewId);
        }

        /// <summary>
        /// Gets all reviews of given product
        /// </summary>
        /// <param name="productId">product id</param>
        /// <returns>all reviews of given product</returns>
        public IEnumerable<ProductReviewDTO> GetAllReviews(int productId = 0)
        {
            return ratingService.GetAllReviewsForProduct(productId);
        } 
        #endregion
    }
}
