﻿using System.Collections.Generic;
using System.Linq;
using AutoMapper;
using BL.DTOs.Categories;
using BL.DTOs.Filters;
using BL.Queries;
using BL.Repositories;
using DAL.Entities;

namespace BL.Services.Categories
{
    /// <summary>
    /// Handles Category CRUD
    /// </summary>
    public class CategoryService : EshopService, ICategoryService
    {
        #region Dependencies

        private readonly CategoryRepository categoryRepository;

        private readonly CategoryListQuery categoryListQuery;

        public CategoryService(CategoryRepository categoryRepository, CategoryListQuery categoryListQuery)
        {
            this.categoryRepository = categoryRepository;
            this.categoryListQuery = categoryListQuery;
        } 

        #endregion

        /// <summary>
        /// Creates new category
        /// </summary>
        /// <param name="categoryDto">category details</param>
        public void CreateCategory(CategoryDTO categoryDto)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var category = Mapper.Map<Category>(categoryDto);
                category.Parent = categoryRepository.GetById(categoryDto.ParentID ?? 0);

                categoryRepository.Insert(category);
                uow.Commit();
            }
        }

        /// <summary>
        /// Updates category according to ID
        /// </summary>
        /// <param name="categoryDto">category details</param>
        public void EditCategory(CategoryDTO categoryDto)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var category = categoryRepository.GetById(categoryDto.ID, categoryEntity => categoryEntity.Parent);
                Mapper.Map(categoryDto, category);
                
                category.Parent = categoryDto.ParentID != null ? 
                    categoryRepository.GetById(categoryDto.ParentID.Value) : null;   
                            
                categoryRepository.Update(category);
                uow.Commit();
            }
        }

        /// <summary>
        /// Removes category according to ID
        /// </summary>
        /// <param name="categoryId">category ID</param>
        public void DeleteCategory(int categoryId)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                categoryRepository.Delete(categoryId);
                uow.Commit();
            }
        }

        /// <summary>
        /// Gets category according to ID
        /// </summary>
        /// <param name="categoryId">category ID</param>
        /// <returns>The category</returns>
        public CategoryDTO GetCategory(int categoryId)
        {
            using (UnitOfWorkProvider.Create())
            {
                var category = categoryRepository.GetById(categoryId);
                return category != null ? Mapper.Map<CategoryDTO>(category) : null;
            }
        }

        /// <summary>
        /// Gets category id that belongs to specified category name
        /// </summary>
        /// <param name="name">category name</param>
        /// <returns>id of category with specified name</returns>
        public int GetProductCategoryIdByName(string name)
        {
            using (UnitOfWorkProvider.Create())
            {
                categoryListQuery.Filter = new CategoryFilter {Name = name};
                var category = categoryListQuery.Execute().SingleOrDefault();
                return category?.ID ?? 0;
            }
        }

        /// <summary>
        /// Gets all parent categories for specified category
        /// </summary>
        /// <param name="categoryId">category id</param>
        /// <returns>all parent categories</returns>
        public IEnumerable<CategoryDTO> GetCategoryPath(int categoryId)
        {
            var list = new List<CategoryDTO>();
            using (UnitOfWorkProvider.Create())
            {
                var category = categoryRepository.GetById(categoryId);
                list.Add(Mapper.Map<CategoryDTO>(category));

                while (category.HasParent)
                {
                    list.Add(Mapper.Map<CategoryDTO>(categoryRepository.GetById(category.Parent.ID)));
                    category = category.Parent;
                }
            }
            return list;
        }

        /// <summary>
        /// Gets all categories
        /// </summary>
        /// <returns>all available categories</returns>
        public IEnumerable<CategoryDTO> ListAllCategories()
        {
            using (UnitOfWorkProvider.Create())
            {
                categoryListQuery.Filter = null;
                return categoryListQuery.Execute() ?? new List<CategoryDTO>();
            }
        }
    }
}
