﻿using System;
using System.Collections.Generic;
using System.Linq;
using AutoMapper;
using BL.DTOs.Coupons;
using BL.DTOs.Filters;
using BL.DTOs.Orders;
using BL.Queries;
using BL.Repositories;
using BL.Utils.Resolvers;
using DAL.Entities;

namespace BL.Services.Checkout
{
    /// <summary>
    /// Handles order checkout related functionality
    /// </summary>
    public class CheckoutService : EshopService, ICheckoutService
    {
        #region Dependencies

        private readonly CouponRepository couponRepository;

        private readonly CouponListQuery couponListQuery;

        private readonly OrderRepository orderRepository;

        private readonly OrderItemRepository orderItemRepository;

        private readonly CustomerRepository customerRepository;

        private readonly ProductRepository productRepository;

        private readonly IPriceCalculatorResolver priceCalculatorResolver;

        public CheckoutService(CouponRepository couponRepository, CouponListQuery couponListQuery, OrderRepository orderRepository, OrderItemRepository orderItemRepository, CustomerRepository customerRepository, ProductRepository productRepository, IPriceCalculatorResolver priceCalculatorResolver)
        {
            this.couponRepository = couponRepository;
            this.couponListQuery = couponListQuery;
            this.orderRepository = orderRepository;
            this.orderItemRepository = orderItemRepository;
            this.customerRepository = customerRepository;
            this.productRepository = productRepository;
            this.priceCalculatorResolver = priceCalculatorResolver;
        }

        #endregion

        #region OrderManagement

        /// <summary>
        /// Persists order together with all related data
        /// </summary>
        /// <param name="createOrderDto">wrapper for order, orderItems, customer and coupon</param>
        /// <param name="confirmationAction">Action to perform when uow successfully performs after commit action</param>
        public void ConfirmOrder(OrderCreateDTO createOrderDto, Action confirmationAction)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                uow.RegisterAfterCommitAction(confirmationAction);

                var order = Mapper.Map<Order>(createOrderDto.Order);

                var customer = customerRepository.GetById(createOrderDto.CustomerId);
                if (customer == null)
                {
                    throw new ArgumentException("CheckoutService - CreateOrder(...) Customer must not be null");
                }
                order.Customer = customer;

                if (createOrderDto.CouponId > 0)
                {
                    var coupon = couponRepository.GetById(createOrderDto.CouponId);
                    order.Coupon = coupon; // coupon activation
                }

                foreach (var orderItemDto in createOrderDto.OrderItems)
                {
                    var orderItem = Mapper.Map<OrderItem>(orderItemDto);
                    orderItem.Product = productRepository.GetById(orderItemDto.Product.ID);
                    orderItem.Order = order;
                    orderItemRepository.Insert(orderItem);
                }

                orderRepository.Insert(order);

                uow.Commit();
            }
        }

        /// <summary>
        /// Gets discount coupon according to promo code
        /// </summary>
        /// <param name="promoCode">Coupon promo code</param>
        /// <param name="customerId">ID of customer who owns the coupon</param>
        /// <returns>Respective discount coupon</returns>
        public CouponDTO GetCoupon(string promoCode, int customerId)
        {
            if (customerId == 0)
            {
                throw new ArgumentException("CheckoutService - GetCoupon(...) customer ID must not equal zero");
            }
            using (UnitOfWorkProvider.Create())
            {
                couponListQuery.Filter = new CouponFilter { IsUsed = false, PromoCode = promoCode, CustomerId = customerId };
                return couponListQuery.Execute().SingleOrDefault();
            }
        }

        /// <summary>
        /// Notifies warehouse about realised order and requests corresponding order shipment
        /// </summary>
        public void RequestOrderDelivery()
        {
            /*  for instance a request to https://www.easypost.com/dhl-express-api 
                can be made, corresponding warehouse receives a notification about
                delivery request and schedules delivery
            */
        } 
        
        #endregion

        #region TotalPriceCalculation

        /// <summary>
        /// Calculates total price for all order items (overall coupon discount is included)
        /// </summary>
        /// <param name="orderItems">all order items</param>
        /// <param name="couponId">coupon id</param>
        /// <returns>Total price for given items</returns>
        public decimal CalculateTotalPrice(ICollection<OrderItemDTO> orderItems, int couponId = 0)
        {
            if (orderItems == null || orderItems.Count == 0)
            {
                return 0;
            }            
            var itemsPrice = GetItemsTotalPrice(orderItems.ToArray());
            return ApplyCouponDiscount(couponId, itemsPrice);
        }

        /// <summary>
        /// Gets total price only for order items
        /// </summary>
        /// <param name="orderItems">all order items</param>
        /// <returns>The total price for all items</returns>
        public decimal GetItemsTotalPrice(params OrderItemDTO[] orderItems)
        {
            decimal itemsPrice = 0;
            foreach (var orderItem in orderItems)
            {
                var calculator = priceCalculatorResolver.Resolve(orderItem.Product.DiscountType);
                itemsPrice += calculator.CalculatePrice(orderItem);
            }
            return itemsPrice;
        }

        /// <summary>
        /// Applies coupon discount on all order items price
        /// </summary>
        /// <param name="couponId">coupon id</param>
        /// <param name="itemsPrice">all order items price</param>
        /// <returns></returns>
        private decimal ApplyCouponDiscount(int couponId, decimal itemsPrice)
        {
            if (couponId == 0)
            {
                return itemsPrice;
            }
            using (UnitOfWorkProvider.Create())
            {
                var coupon = couponRepository.GetById(couponId);
                if (coupon != null && coupon.Order == null)
                {
                    return itemsPrice * (decimal)(1 - coupon.DiscountPercentage / 100.0);
                }

            }
            return itemsPrice;
        } 
        
        #endregion
    }
}
