﻿using System;
using System.Collections.Generic;
using System.Linq;
using AutoMapper;
using BL.DTOs.Coupons;
using BL.DTOs.Customers;
using BL.DTOs.Filters;
using BL.DTOs.Orders;
using BL.Queries;
using BL.Repositories;
using BL.Repositories.UserAccount;
using DAL.Entities;
using Riganti.Utils.Infrastructure.Core;

namespace BL.Services.Customers
{
    public class CustomerService : EshopService, ICustomerService
    {
        #region PageSizeConstants

        private const int OrdersPageSize = 10;

        private const int CouponsPageSize = 5; 
        
        public int CustomersPageSize => 20;

        #endregion

        #region Dependencies

        private readonly UserAccountRepository userRepository;

        private readonly CustomerRepository customerRepository;

        private readonly OrderListQuery orderListQuery;

        private readonly CustomerListQuery customerListQuery;

        private readonly CustomerAccordingToEmailQuery customerAccordingToEmailQuery;

        private readonly CouponListQuery couponListQuery;

        public CustomerService(UserAccountRepository userRepository, CustomerRepository customerRepository, OrderListQuery orderListQuery, CustomerListQuery customerListQuery, CustomerAccordingToEmailQuery customerAccordingToEmailQuery, CouponListQuery couponListQuery)
        {
            this.userRepository = userRepository;
            this.customerRepository = customerRepository;
            this.orderListQuery = orderListQuery;
            this.customerListQuery = customerListQuery;
            this.customerAccordingToEmailQuery = customerAccordingToEmailQuery;
            this.couponListQuery = couponListQuery;
        }

        #endregion

        /// <summary>
        /// Gets customer according to ID
        /// </summary>
        /// <param name="customerId">customer ID</param>
        /// <returns>The customer</returns>
        public CustomerDTO GetCustomer(int customerId)
        {
            using (UnitOfWorkProvider.Create())
            {
                var customer = customerRepository.GetById(customerId, c => c.Account);
                return customer != null ? Mapper.Map<CustomerDTO>(customer) : null;
            }
        }

        /// <summary>
        /// Creates new customer (user account must be created first)
        /// </summary>
        /// <param name="userAccountId">Customer user account ID</param>
        public void CreateCustomer(Guid userAccountId)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var customerAccount = userRepository.GetById(userAccountId);

                var customer = new Customer {Account = customerAccount};

                customerRepository.Insert(customer);

                uow.Commit();
            }
        }

        /// <summary>
        /// Updates customer according to ID
        /// </summary>
        /// <param name="customerDto">customer details</param>
        public void EditCustomer(CustomerDTO customerDto)
        {
            // edit corresponding user account details or manage entities in customer class preferably in some CustomerEditDTO

            using (var uow = UnitOfWorkProvider.Create())
            {
                var customer = customerRepository.GetById(customerDto.ID);
                Mapper.Map(customerDto, customer);

                customerRepository.Update(customer);

                uow.Commit();
            }
        }

        /// <summary>
        /// Gets customer with given email address
        /// </summary>
        /// <param name="email">email</param>
        /// <returns>Customer with given email address</returns>
        public CustomerDTO GetCustomerAccordingToEmail(string email)
        {
            using (UnitOfWorkProvider.Create())
            {
                customerAccordingToEmailQuery.Email = email;
                return customerAccordingToEmailQuery.Execute().SingleOrDefault();
            }
        }

        /// <summary>
        /// Gets customers according to requested page
        /// </summary>
        /// <returns>Customers page</returns>
        public CustomerListQueryResultDTO ListCustomers(int requiredPage)
        {
            using (UnitOfWorkProvider.Create())
            {
                var query = GetCustomerQuery();
                query.Skip = Math.Max(0, requiredPage - 1) * CustomersPageSize;
                query.Take = CustomersPageSize;

                query.AddSortCriteria(customer => customer.LastName);

                return new CustomerListQueryResultDTO
                {
                    RequestedPage = requiredPage,
                    TotalResultCount = query.GetTotalRowCount(),
                    ResultsPage = query.Execute()
                };
            }
        }

        /// <summary>
        /// Gets all orders according to filter and required page
        /// </summary>
        /// <param name="filter">order filter</param>
        /// <param name="requiredPage">page to show</param>
        /// <returns>All orders</returns>
        public IEnumerable<OrderDTO> ListOrders(OrderFilter filter, int requiredPage = 1)
        {
            using (UnitOfWorkProvider.Create())
            {
                orderListQuery.Filter = filter;
                orderListQuery.Skip = Math.Max(0, requiredPage - 1) * OrdersPageSize;
                orderListQuery.Take = OrdersPageSize;

                var sortOrder = filter.SortAscending ? SortDirection.Ascending : SortDirection.Descending;
                orderListQuery.AddSortCriteria("Issued", sortOrder);

                return orderListQuery.Execute();
            }
        }

        /// <summary>
        /// Gets all coupons according to filter and required page
        /// </summary>
        /// <param name="filter">coupon filter</param>
        /// <param name="requiredPage">page to show</param>
        /// <returns>All coupons</returns>
        public IEnumerable<CouponDTO> ListCoupons(CouponFilter filter, int requiredPage = 1)
        {
            using (UnitOfWorkProvider.Create())
            {
                couponListQuery.Filter = filter;
                couponListQuery.Skip = Math.Max(0, requiredPage - 1) * CouponsPageSize;
                couponListQuery.Take = CouponsPageSize;

                var sortOrder = filter.SortAscending ? SortDirection.Ascending : SortDirection.Descending;
                couponListQuery.AddSortCriteria("Issued", sortOrder);

                return couponListQuery.Execute();
            }
        }

        /// <summary>
        /// Configures customers list query
        /// </summary>
        /// <returns>configured query</returns>
        private IQuery<CustomerDTO> GetCustomerQuery()
        {
            var query = customerListQuery;
            query.ClearSortCriterias();
            return query;
        }
    }
}
