﻿using System;
using System.Linq;
using AutoMapper;
using BL.DTOs.Filters;
using BL.DTOs.Products;
using BL.Queries;
using BL.Repositories;
using DAL.Entities;
using Riganti.Utils.Infrastructure.Core;

namespace BL.Services.Products
{
    /// <summary>
    /// Handles Product CRUD and querying (with paging support)
    /// </summary>
    public class ProductService : EshopService, IProductService
    {
        public int ProductsPageSize => 9;

        #region Dependencies

        private readonly CategoryRepository categoryRepository;

        private readonly ProductReviewRepository productReviewRepository;

        private readonly ProductRepository productRepository;

        private readonly ProductListQuery productListQuery;

        public ProductService(CategoryRepository categoryRepository, ProductReviewRepository productReviewRepository, ProductRepository productRepository, ProductListQuery productListQuery)
        {
            this.categoryRepository = categoryRepository;
            this.productReviewRepository = productReviewRepository;
            this.productRepository = productRepository;
            this.productListQuery = productListQuery;
        }
        
        #endregion

        /// <summary>
        /// Get product according to ID
        /// </summary>
        /// <param name="productId">ID of the product</param>
        /// <returns>The product with given ID, null otherwise</returns>
        public ProductDTO GetProduct(int productId)
        {
            using (UnitOfWorkProvider.Create())
            {
                var product = productRepository.GetById(productId);
                return product != null ? Mapper.Map<ProductDTO>(product) : null;    
            }
        }

        /// <summary>
        /// Gets products according to filter and required page
        /// </summary>
        /// <param name="filter">products filter</param>
        /// <param name="requiredPage">page to show</param>
        /// <returns></returns>
        public ProductListQueryResultDTO ListProducts(ProductFilter filter, int requiredPage = 1)
        {
            using (UnitOfWorkProvider.Create())
            {
                var query = GetQuery(filter);
                query.Skip = Math.Max(0, requiredPage - 1) * ProductsPageSize;
                query.Take = ProductsPageSize;

                var sortOrder = filter.SortAscending ? SortDirection.Ascending : SortDirection.Descending;
                query.AddSortCriteria(filter.SortCriteria.ToString(), sortOrder);

                return new ProductListQueryResultDTO
                {
                    RequestedPage = requiredPage,
                    TotalResultCount = query.GetTotalRowCount(),
                    ResultsPage = query.Execute(),
                    Filter = filter
                };                
            }
        }

        /// <summary>
        /// Creates new product
        /// </summary>
        /// <param name="productDto">Product details</param>
        /// <param name="categoryId">Product category ID</param>
        public void CreateProduct(ProductDTO productDto, int categoryId)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var product = Mapper.Map<Product>(productDto);
                product.Category = GetProductCategory(categoryId);
                product.Stock = new Stock()
                {
                    Product = product,
                    StoredUnits = 00
                };
                productRepository.Insert(product);
                uow.Commit();
            }
        }

        /// <summary>
        /// Updates product
        /// </summary>
        /// <param name="productDto">Product details</param>
        /// <param name="categoryId">Product category ID</param>
        /// <param name="productReviewIds">Product reviews IDs</param>
        public void EditProduct(ProductDTO productDto, int categoryId, params int[] productReviewIds)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var product = productRepository.GetById(productDto.ID, p => p.Stock);
                Mapper.Map(productDto, product);

                product.Category = GetProductCategory(categoryId);

                if (productReviewIds != null && productReviewIds.Any())
                {
                    var productReviews = productReviewRepository.GetByIds(productReviewIds);
                    product.ProductReviews.RemoveAll(review => !productReviews.Contains(review));
                    product.ProductReviews.AddRange(
                        productReviews.Where(review => !product.ProductReviews.Contains(review)));
                }
                else
                {
                    product.ProductReviews.Clear();
                }

                productRepository.Update(product);
                uow.Commit();
            }
        }

        /// <summary>
        /// Deletes product with given ID
        /// </summary>
        /// <param name="productId">ID of the product to delete</param>
        public void DeleteProduct(int productId)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                productRepository.Delete(productId);
                uow.Commit();
            }
        }
        
        /// <summary>
        /// Gets required category (throws exception if category with given id does not exists)
        /// </summary>
        /// <param name="categoryId">Id of the category</param>
        /// <returns>category according to id</returns>
        private Category GetProductCategory(int categoryId)
        {
            var category = categoryRepository.GetById(categoryId);
            if (category == null)
            {
                throw new NullReferenceException("Product service - CreateProduct(...) category cant be null");
            }
            return category;              
        }

        /// <summary>
        /// Configures products list query
        /// </summary>
        /// <param name="filter">product filter</param>
        /// <returns>configured query</returns>
        private IQuery<ProductDTO> GetQuery(ProductFilter filter)
        {
            var query = productListQuery;
            query.ClearSortCriterias();
            query.Filter = filter;
            return query;
        }
    }
}
