﻿using System;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Net;
using System.Net.Mail;
using BL.DTOs.Coupons;
using BL.DTOs.Filters;
using BL.Queries;
using BL.Repositories;
using DAL.Entities;

namespace BL.Services.Promote
{
    /// <summary>
    /// Handles e-shop promo offers (discount coupons)
    /// </summary>
    public class PromoteService : EshopService, IPromoteService
    {
        #region Promo related constants

        private const string PromoCodeChars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";

        private const int BirthdayDiscountPercentage = 20;

        private const int LoyaltyDiscountPercentage = 5;

        /// <summary>
        /// number of days before birthday when customer is elegible to receive birthday coupon
        /// </summary>
        private const int NumberOfDaysBeforeBirthday = 30;
        
        #endregion

        #region E-mail

        private const string Subject = "<p>Demo E-shop discount</p>";

        private const string Greeting = "<p>Dear customer:</p>";

        private const string LoyaltyMsg =
            "<p>We are very pleased by your loyalty, therefore we would like to thank you with following discount code:</p>";

        private const string BirthdayMsg =
            "<p>We wish you happy birthday and would like to offer you following discount code as your birthday gift:</p>";

        private const string FinalParagraph =
            "<p>You can use above code exactly once to any of your order.<p>Sincerely<br>Demo E-shop team";

        #endregion

        #region Dependencies

        private readonly CouponRepository couponRepository;

        private readonly CustomerRepository customerRepository;

        private readonly CouponListQuery couponListQuery; 
        
        public PromoteService(CouponRepository couponRepository, CustomerRepository customerRepository, CouponListQuery couponListQuery)
        {
            this.couponRepository = couponRepository;
            this.customerRepository = customerRepository;
            this.couponListQuery = couponListQuery;
        }

        #endregion

        /// <summary>
        /// Creates discount coupon for specified customer
        /// </summary>
        /// <param name="customerId">Id of customer who will be given the coupon</param>
        /// <param name="promoCode">Optional promo code</param>
        /// <param name="asBirthdayCoupon">create as birthday coupon</param>
        public string CreateCoupon(int customerId, string promoCode = null, bool asBirthdayCoupon = false)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var customer = customerRepository.GetById(customerId);
                if (customer == null)
                {
                    throw new NullReferenceException($"Promote service - CreateCoupon(...) customer cant be null");
                }
                var coupon = couponRepository.InitializeNew();
                coupon.Customer = customer;
                coupon.Issued = DateTime.UtcNow;
                coupon.PromoCode = promoCode ?? GenerateCouponCode();
                if (asBirthdayCoupon && !CheckCustomerHasBirthdayCoupon(customerId))
                {
                    coupon.DiscountPercentage = BirthdayDiscountPercentage;
                }
                else
                {
                    coupon.DiscountPercentage = LoyaltyDiscountPercentage;
                }

                couponRepository.Insert(coupon);

                uow.Commit();
                return coupon.PromoCode;
            }
        }

        /// <summary>
        /// Sends birthday coupon (in fact just the promo code) to 
        /// corresponding customer if and only if all conditions are satisfied
        /// </summary>
        /// <param name="customerId">ID of coupon receiver</param>
        /// <returns>true if coupon was sent</returns>
        public bool SendBirthDayCouponIfPossible(int customerId)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var customer = customerRepository.GetById(customerId);
                if (customer == null)
                {
                    throw new NullReferenceException("Promote service -  CheckSuitabilityForBirthDayCoupon(...) customer cant be null");
                }
                var isSuitable = DetermineSuitabilityForBirthdayCoupon(customer);
                if (isSuitable)
                {
                    var promoCode = CreateCoupon(customerId, null, true);
                    SendPromoCode(customer.Account.Email, promoCode, false);
                    uow.Commit();
                }  
                return isSuitable;            
            }
        }

        /// <summary>
        /// Sends loyalty coupon(in fact just the promo code) to
        /// corresponding customer if and only if all conditions are satisfied
        /// </summary>
        /// <param name="customerId">ID of coupon receiver</param>
        /// <returns>true if coupon was sent</returns>
        public bool SendLoyaltyCouponIfPossible(int customerId)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var customer = customerRepository.GetById(customerId, c => c.Orders);
                if (customer != null && customer.Orders.Count > 2)
                {
                    var promoCode = CreateCoupon(customerId);
                    SendPromoCode(customer.Account.Email, promoCode);
                    uow.Commit();
                    return true;
                }
                return false;
            }
        }

        /// <summary>
        /// Sends promo code discount
        /// </summary>
        /// <param name="customerEmail">mail to send promo code to</param>
        /// <param name="code">the promo code</param>
        private void SendPromoCode(string customerEmail, string code, bool isLoyaltyDiscount = true)
        {
            if (string.IsNullOrEmpty(customerEmail) || !new EmailAddressAttribute().IsValid(customerEmail))
            {
                throw new ArgumentException("Promote service - SendPromoCode(...) customerEmail must be valid email address");
            }

            if (string.IsNullOrEmpty(code) || !code.All(char.IsLetterOrDigit))
            {
                throw new ArgumentException("Promote service - SendPromoCode(...) code must consists only of alpha numeric characters");
            }

            var introMsg = isLoyaltyDiscount ? LoyaltyMsg : BirthdayMsg;

            var message = new MailMessage();
            message.To.Add(new MailAddress(customerEmail));  // replace with valid value 
            message.From = new MailAddress("pv226demoEshop@gmail.com");  // replace with valid value
            message.Subject = Subject;
            message.Body = Greeting + introMsg + SpanCode(code) + FinalParagraph;
            message.IsBodyHtml = true;

            SendMail(message);
        }

        /// <summary>
        /// Checks whether customer is eligible to receive B-day coupon
        /// </summary>
        /// <param name="customer">Customer to check</param>
        /// <returns>true if customer is eligible to receive B-day coupon</returns>
        private bool DetermineSuitabilityForBirthdayCoupon(Customer customer)
        {
            var validationDate = GetBirthdayValidationDate(customer);
            return validationDate.AddDays(-1*NumberOfDaysBeforeBirthday) <= DateTime.UtcNow && DateTime.UtcNow <= validationDate;
        }

        /// <summary>
        /// Sends email message with discount coupon
        /// </summary>
        /// <param name="message">the mail message</param>
        private static void SendMail(MailMessage message)
        {
            using (var client = new SmtpClient())
            {
                client.Host = "smtp.gmail.com";
                client.Port = 587;
                client.EnableSsl = true;
                client.DeliveryMethod = SmtpDeliveryMethod.Network;
                client.UseDefaultCredentials = false;
                client.Credentials = new NetworkCredential
                {
                    UserName = "pv226demoEshop@gmail.com", // replace with valid value
                    Password = "secretPa$$" // replace with valid value
                };

                client.Send(message);
            }
        }

        /// <summary>
        /// Checks if customer already has birthday coupon
        /// </summary>
        /// <param name="customerId">customer to check</param>
        /// <returns>true if customer already has birthday coupon</returns>
        private bool CheckCustomerHasBirthdayCoupon(int customerId)
        {
            using (UnitOfWorkProvider.Create())
            {
                var customer = customerRepository.GetById(customerId);
                if (customer == null)
                {
                    throw new NullReferenceException("Promote service - CheckCustomerHasBirthdayCoupon(...) customer cant be null");
                }

                couponListQuery.Filter = new CouponFilter {CustomerId = customerId, IsUsed = false};
                var availableCoupons = couponListQuery.Execute();
                return availableCoupons.Any(coupon => CheckCouponIsForBirthday(customer, coupon));
            }
        }

        /// <summary>
        /// Checks whether discount coupon can be used for birthday
        /// </summary>
        /// <param name="customer">customer who owns the discount coupon</param>
        /// <param name="coupon">the discount coupon to check</param>
        /// <returns>true if given coupon is for birthday</returns>
        private bool CheckCouponIsForBirthday(Customer customer, CouponDTO coupon)
        {
            var validationDate = GetBirthdayValidationDate(customer);
            return validationDate.AddDays(-1 * NumberOfDaysBeforeBirthday) >= coupon.Issued &&
                    coupon.Issued <= validationDate &&
                    coupon.DiscountPercentage == BirthdayDiscountPercentage;
        }

        /// <summary>
        /// Gets birthday validation date (customer birthday date with current year)
        /// </summary>
        /// <param name="customer">customer for which the validation date will be created</param>
        /// <returns>validation date for corresponding customer</returns>
        private DateTime GetBirthdayValidationDate(Customer customer)
        {
            var yearsToAdd = DateTime.UtcNow.Year - customer.Account.BirthDate.Year;
            return customer.Account.BirthDate.AddYears(yearsToAdd);
        }

        /// <summary>
        /// creates unique 8-char discount code
        /// </summary>
        /// <returns>discount code</returns>
        private string GenerateCouponCode()
        {
            var stringChars = new char[8];
            var rand = new Random();

            for (var i = 0; i < stringChars.Length; i++)
            {
                stringChars[i] = PromoCodeChars[rand.Next(PromoCodeChars.Length)];
            }
            return new string(stringChars);
        }

        /// <summary>
        /// Formats the promo code by HTML tags 
        /// </summary>
        /// <param name="code">The promo code</param>
        /// <returns>Promo code in HTML format</returns>
        private string SpanCode(string code) => "<p><b>" + code + "</b><p>";
    }
}
