﻿using System;
using System.Collections.Generic;
using AutoMapper;
using BL.DTOs.Filters;
using BL.DTOs.Reviews;
using BL.Queries;
using BL.Repositories;
using DAL.Entities;

namespace BL.Services.Rating
{
    public class RatingService : EshopService, IRatingService
    {
        #region Dependencies

        private ProductReviewListQuery productReviewListQuery;

        private ProductReviewRepository productReviewRepository;

        private ProductRepository productRepository;

        public RatingService(ProductReviewListQuery productReviewListQuery, ProductReviewRepository productReviewRepository, ProductRepository productRepository)
        {
            this.productReviewListQuery = productReviewListQuery;
            this.productReviewRepository = productReviewRepository;
            this.productRepository = productRepository;
        } 

        #endregion

        /// <summary>
        /// Gets all reviews of given product
        /// </summary>
        /// <param name="productId">product id</param>
        /// <returns>all reviews of given product</returns>
        public IEnumerable<ProductReviewDTO> GetAllReviewsForProduct(int productId)
        {
            using (UnitOfWorkProvider.Create())
            {
                productReviewListQuery.Filter = new ProductReviewFilter {ProductId = productId};
                return productReviewListQuery.Execute();
            }
        }

        /// <summary>
        /// Adds review
        /// </summary>
        /// <param name="reviewDto">Review details</param>
        public void AddReview(ProductReviewCreateDTO reviewDto)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var review = Mapper.Map<ProductReview>(reviewDto);
                review.Product = GetReviewProduct(reviewDto.ProductID);
                productReviewRepository.Insert(review);
                uow.Commit();
            }
        }

        /// <summary>
        /// Updates review
        /// </summary>
        /// <param name="reviewDto">Update details</param>
        public void EditReview(ProductReviewEditDTO reviewDto)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var review = productReviewRepository.GetById(reviewDto.ID, productReview => productReview.Product);
                if (review == null)
                {
                    throw new NullReferenceException("Rating service - EditReview(...) review cant be null");
                }
                Mapper.Map(reviewDto, review);                
                productReviewRepository.Update(review);
                uow.Commit();
            }
        }

        /// <summary>
        /// Deletes review
        /// </summary>
        /// <param name="id">ID of review to delete</param>
        public void DeleteReview(int id)
        {
            using (var uow = UnitOfWorkProvider.Create())
            {
                var review = productReviewRepository.GetById(id, productReview => productReview.Product);
                if (review == null)
                {
                    throw new NullReferenceException("Rating service - DeleteReview(...) review cant be null");
                }
                productReviewRepository.Delete(review);
                uow.Commit();
            }
        }

        /// <summary>
        /// Gets product for corresponding review according to ID
        /// </summary>
        /// <param name="productId">The product id</param>
        /// <returns>product according to ID</returns>
        private Product GetReviewProduct(int productId)
        {
            var product = productRepository.GetById(productId);
            if (product == null)
            {
                throw new NullReferenceException("Rating service - GetReviewProduct(...) product cant be null");
            }
            return product;
        }
    }
}
