﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Text;
using BL.DTOs.Stocks;
using BL.Repositories;
using DAL.Entities;

namespace BL.Services.Stocks
{
    /// <summary>
    /// Manages stocks (orders, dispatch and reservations) 
    /// </summary>
    public class StockService : EshopService, IStockService
    {
        private static readonly object AvailableUnitsLock = new object();

        /// <summary>
        /// Preserves number of reserved units per customer for each stock
        /// Key is id of the corresponding stock
        /// </summary>
        private static readonly ConcurrentDictionary<int, IList<StockReservationDTO>> ReservedStockUnits = new ConcurrentDictionary<int, IList<StockReservationDTO>>();

        #region Dependencies

        private readonly StockRepository stockRepository;

        public StockService(StockRepository stockRepository)
        {
            this.stockRepository = stockRepository;
        } 

        #endregion

        /// <summary>
        /// Performs stock reservation in order to ensure
        /// simultaneous orders can be satisfied
        /// </summary>
        /// <param name="stockReservation">Stock reservation details containing 
        /// customer reserving the stock, amount, stockID and expiration</param>
        /// <returns>true if reservation is successfull, otherwise false</returns>
        public bool ReserveStock(StockReservationDTO stockReservation)
        {
            PerformExpiredReservationRelease();
            CheckStockReservationIsValid(stockReservation);            
            EnsureDictionaryRecordIsInitialized(stockReservation.StockId);

            lock (AvailableUnitsLock)
            {
                return ReserveStockWithinLockCore(stockReservation);
            }
        }

        /// <summary>
        /// Performs stock release 
        /// (typically when user confirms or cancels the order)
        /// </summary>
        /// <param name="customerId">ID of customer who reserved the stock</param>
        /// <param name="stockIds">IDs of the stock to release</param>
        public void ReleaseStockReservations(int customerId, params int[] stockIds)
        {
            if (customerId == 0)
            {
                return;
            }

            Func<StockReservationDTO, bool> deletePredicate;
            var cancelAllCustomerReservations = stockIds.Length == 0;
           
            if (cancelAllCustomerReservations)
            {
                deletePredicate = reservation => reservation.CustomerId == customerId;
            }
            else
            {
                deletePredicate = reservation => stockIds.Contains(reservation.StockId) && reservation.CustomerId == customerId;
            }

            ReleaseStockReservationsCore(deletePredicate);
        }

        /// <summary>
        /// Performs stock expedition by decreasing stocks 
        /// stored units and releasing corresponding reservations
        /// </summary>
        /// <param name="customerId">Id of customer whose reserved stock should be dispatched</param>
        public void DispatchStock(int customerId)
        {
            lock (AvailableUnitsLock)
            {
                using (var uow = UnitOfWorkProvider.Create())
                {
                    var allCustomerReservations = new List<StockReservationDTO>();

                    foreach (var reservationList in ReservedStockUnits.Values
                        .Where(list => list.Any(item => item.CustomerId == customerId))
                        .ToList())
                    {
                        foreach (var reservation in reservationList
                            .Where(reservation => reservation.CustomerId == customerId).ToList())
                        {
                            allCustomerReservations.Add(reservation);
                            reservationList.Remove(reservation);
                        }
                    }

                    foreach (var customerReservation in allCustomerReservations)
                    {
                        Stock stock;
                        if (TryGetStockWithinLock(customerReservation.StockId, out stock))
                        {
                            stock.StoredUnits = Math.Max(0, stock.StoredUnits - customerReservation.ReservedAmount);
                        }                  
                    }
                    uow.Commit();
                }
            }
        }

        /// <summary>
        /// Performs order of required stock from distributor
        /// </summary>
        /// <param name="stockId">id of stock to order</param>
        /// <param name="quantity">Number of units to order</param>
        public void OrderStockFromDistributor(int stockId, int quantity)
        {
            lock (AvailableUnitsLock)
            {
                using (var uow = UnitOfWorkProvider.Create())
                {
                    var stock = stockRepository.GetById(stockId, s => s.Product);
                    if (stock == null)
                    {
                        throw new ArgumentException("StockService - OrderStockFromDistributor(...): stock ID does not belong to any stored stock");
                    }
                    if (quantity < 1)
                    {
                        throw new ArgumentException("StockService - OrderStockFromDistributor(...): quantity must be greater than zero");
                    }
                    stock.StoredUnits += quantity;
                    stockRepository.Update(stock);
                    uow.Commit();
                }
            }
        }

        /// <summary>
        /// Gets number of currently available (free)
        /// units for stock belonging to product with specified ID
        /// </summary>
        /// <param name="stockId">ID of the stock for which stock availability should be checked</param>
        /// <returns>Number of free units for this product</returns>
        public int GetCurrentlyAvailableUnits(int stockId)
        {
            PerformExpiredReservationRelease();
            EnsureDictionaryRecordIsInitialized(stockId);
            lock (AvailableUnitsLock)
            {
                return GetCurrentlyAvailableUnitsWithinLock(stockId);
            }
        }


        /// <summary>
        /// Gets number of currently available (free)
        /// units for stock belonging to product with specified ID
        /// </summary>
        /// <param name="stockId">ID of the stock for which stock availability should be checked</param>
        /// <returns>Number of free units for this product</returns>
        private int GetCurrentlyAvailableUnitsWithinLock(int stockId)
        {
            Stock stock;
            if (!TryGetStockWithinLock(stockId, out stock))
            {
                return 0;
            }
            var storedUnits = stock.StoredUnits;
            var reservedItems = ReservedStockUnits[stockId].ToList();
            var reservedUnits = reservedItems.Sum(item => item.ReservedAmount);
            return storedUnits - reservedUnits;
        }

        /// <summary>
        /// Gets stock according to ID.
        /// This method should be called within synchronised context.
        /// </summary>
        /// <param name="stockId">id of the stock</param>
        /// <param name="stock">retrieved stock</param>
        /// <returns>true if stock retrieval was successfull</returns>
        private bool TryGetStockWithinLock(int stockId, out Stock stock)
        {
            using (UnitOfWorkProvider.Create())
            {
                stock = stockRepository.GetById(stockId, s => s.Product);
            }
            return stock != null;
        }

        /// <summary>
        /// Performs stock the actual stock reservation in order to ensure
        /// simultaneous orders can be satisfied properly
        /// </summary>
        /// <param name="stockReservation">Stock reservation details containing 
        /// customer reserving the stock, amount, stockID and expiration</param>
        /// <returns>true if reservation is successfull, otherwise false</returns>
        private bool ReserveStockWithinLockCore(StockReservationDTO stockReservation)
        {
            var freeUnits = GetCurrentlyAvailableUnitsWithinLock(stockReservation.StockId);
            
            var customerReservation = ReservedStockUnits[stockReservation.StockId]
                                        .FirstOrDefault(item => item.CustomerId == stockReservation.CustomerId);

            if (customerReservation != null)
            {
                // refresh expiration time
                customerReservation.Expiration = stockReservation.Expiration;

                if (stockReservation.ReservedAmount == customerReservation.ReservedAmount)
                {
                    return true;
                }

                if (stockReservation.ReservedAmount <= freeUnits + customerReservation.ReservedAmount)
                {
                    customerReservation.ReservedAmount = stockReservation.ReservedAmount;
                    return true;
                }
            }
            else
            {
                if (stockReservation.ReservedAmount <= freeUnits)
                {
                    ReservedStockUnits[stockReservation.StockId].Add(stockReservation);
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Performs validation of stock reservation.
        /// </summary>
        /// <param name="stockReservation">Stock reservation to validate</param>
        private static void CheckStockReservationIsValid(StockReservationDTO stockReservation)
        {
            var context = new ValidationContext(stockReservation);
            var results = new List<ValidationResult>();

            var isValid = Validator.TryValidateObject(stockReservation, context, results, true);
            if (isValid)
            {
                return;
            }
            var msgBuilder = new StringBuilder();
            foreach (var result in results)
            {
                msgBuilder.Append(result.ErrorMessage);
            }
            throw new ArgumentException($"StockService - stock reservation DTO is not valid ({msgBuilder}).");
        }

        /// <summary>
        /// Creates record within ReservedStockUnits if it does not exist (for the given stockId) yet.
        /// </summary>
        /// <param name="stockId">ID of the stock to check.</param>
        private static void EnsureDictionaryRecordIsInitialized(int stockId)
        {
            lock (AvailableUnitsLock)
            {
                if (!ReservedStockUnits.ContainsKey(stockId))
                {
                    ReservedStockUnits.AddOrUpdate(stockId, new List<StockReservationDTO>(), (key, values) => values);
                }               
            }
        }

        /// <summary>
        /// Releases all expired reservation from ReservedStockUnits.
        /// </summary>
        private static void PerformExpiredReservationRelease()
        {
            ReleaseStockReservationsCore(reservation => DateTime.Now > reservation.Expiration);
        }

        /// <summary>
        /// Releases all expired reservation from ReservedStockUnits based on given predicate.
        /// </summary>
        /// <param name="deletePredicate">Predicate which decides whether respective reservation should be removed.</param>
        private static void ReleaseStockReservationsCore(Func<StockReservationDTO, bool> deletePredicate)
        {
            lock (AvailableUnitsLock)
            {
                foreach (var reservationList in ReservedStockUnits.Values)
                {
                    foreach (var reservation in reservationList.ToList())
                    {
                        if (deletePredicate(reservation))
                        {
                            reservationList.Remove(reservation);
                        }
                    }
                }
            }
        }
    }
}
