﻿using System;
using System.Collections.Generic;
using System.Linq;
using BL;
using BL.Bootstrap;
using BL.DTOs.Categories;
using BL.DTOs.Filters;
using BL.DTOs.Orders;
using BL.DTOs.Products;
using BL.DTOs.Reviews;
using BL.DTOs.Stocks;
using BL.Services.Categories;
using BL.Services.Checkout;
using BL.Services.Customers;
using BL.Services.Products;
using BL.Services.Promote;
using BL.Services.Rating;
using BL.Services.Sales;
using BL.Services.Stocks;
using BL.Utils;
using BL.Utils.DataRepresentation;
using BL.Utils.Enums;
using BrockAllen.MembershipReboot;
using Castle.Windsor;

namespace ConsoleApp
{
    /// <summary>
    /// Main test class for BL functionality
    /// </summary>
    public class Program
    {
        private static ICategoryService categoryService;
        private static IProductService productService;
        private static ICustomerService customerService;
        private static IStockService stockService;
        private static ICheckoutService checkoutService;
        private static ISalesService salesService;
        private static IRatingService ratingService;
        private static IPromoteService promoteService;
        private static UserAccountService<Riganti.Utils.Infrastructure.EntityFramework.UserAccount> userAccountService;

        private static int category1Id;
        private static int category2Id;
        private static int category3Id;

        private static readonly IWindsorContainer Container = new WindsorContainer();

        static void Main(string[] args)
        {
            // configure DI
            InitializeWindsorContainerAndMapper();

            // do the testing
            TestStocksService();
            TestCategoryService();
            TestProductService();
            TestRatingService();
            TestUserAccountService();
            TestCheckoutService();
            TestPromoteService();
        }
        
        private static void InitializeWindsorContainerAndMapper()
        {
            Container.Install(new BussinessLayerInstaller());

            MappingInit.ConfigureMapping();
        }

        private static void TestCategoryService()
        {
            categoryService = Container.Resolve<ICategoryService>();

            //Create
            categoryService.CreateCategory(new CategoryDTO { Name = "Electronics", ParentID = null });
            category1Id = categoryService.GetProductCategoryIdByName("Electronics");
            categoryService.CreateCategory(new CategoryDTO { Name = "Telephones", ParentID = null });
            category2Id = categoryService.GetProductCategoryIdByName("Telephones");
            categoryService.CreateCategory(new CategoryDTO { Name = "Smart phones", ParentID = category2Id });
            category3Id = categoryService.GetProductCategoryIdByName("Smart phones");

            //Rename and Change parent
            categoryService.EditCategory(new CategoryDTO { ID = category2Id, Name = "Phones", ParentID = category1Id });

            //GetCategoryPath
            var list = categoryService.GetCategoryPath(category3Id);
            Console.WriteLine(list.Count() == 3 ? "CategoryService - Test01 - OK" : "CategoryService - Test01 - FAIL");

            var categories = categoryService.ListAllCategories();

            //Delete
            categoryService.DeleteCategory(category3Id);
        }

        private static void TestProductService()
        {
            productService = Container.Resolve<IProductService>();

            //Create
            productService.CreateProduct(new ProductDTO
            {
                Description =
                    "Designed with all the features you love, the Samsung Galaxy P5 is the smartphone you’ve been waiting for. Watching a movie or reading a book is more enjoyable on the large 5.5 HD Super AMOLED display.And while the 13MP main camera captures clearer photos, the 5MP front camera gives you more flattering selfies in any light. Best of all, the long-lasting battery means the Samsung Galaxy P5 keeps up with you.",
                Name = "Samsung Galaxy P5",
                Price = 7420
            }, category2Id);
            productService.CreateProduct(new ProductDTO
            {
                Name = "Lumia 550",
                Description = "Great",
                Price = 3000
            }, category2Id);
            productService.CreateProduct(new ProductDTO
            {
                Name = "Huawei P8",
                Description = "Nice",
                Price = 6200
            }, category2Id);

            // Getting and listing
            var firstProduct = productService.GetProduct(1);
            Console.WriteLine(firstProduct != null ? "ProductService - Test01 - OK" : "ProductService - Test01 - FAIL");
            
            var samsungs = productService.ListProducts(new ProductFilter
            {
                ProductName = "Samsung"
            });

            var cheaps = productService.ListProducts(new ProductFilter
            {
                PriceRange = new Range<decimal> { Minimum = 1000, Maximum = 6500},
                SortCriteria = ProductSortCriteria.Price,
                SortAscending = true
            });

            var expensives = productService.ListProducts(new ProductFilter
            {
                PriceRange = new Range<decimal> { Minimum = 1000, Maximum = 6500 },
                SortCriteria = ProductSortCriteria.Price,
                SortAscending = false
            });
        }

        private static void TestUserAccountService()
        {
            userAccountService = Container.Resolve<UserAccountService<Riganti.Utils.Infrastructure.EntityFramework.UserAccount>>();
            customerService = Container.Resolve<ICustomerService>();
           
            var guid = userAccountService.CreateAccount("Martin", "Password", "410452@mail.muni.cz");
            customerService.CreateCustomer(guid.ID);

            Riganti.Utils.Infrastructure.EntityFramework.UserAccount account;
            AuthenticationFailureCode code;

            var authenticated = userAccountService.Authenticate("410452@mail.muni.cz", "Passwort", out account, out code);
            Console.WriteLine(authenticated == false ? "UserAccountService - Test01 - OK" : "UserAccountService - Test01 - FAIL");

            var authenticated2 = userAccountService.Authenticate("410452@mail.muni.cz", "Password", out account, out code);
            Console.WriteLine(authenticated2 ? "UserAccountService - Test02 - OK" : "UserAccountService - Test02 - FAIL");
        }

        private static void TestStocksService()
        {
            var galaxyJ7StockId = 1;
            var martinCustomerId = 1;

            stockService = Container.Resolve<IStockService>();
            
            var samsungGalaxyJ7Reservation = new StockReservationDTO
            {
                CustomerId = martinCustomerId,
                Expiration = DateTime.Now.AddMinutes(30),
                StockId = galaxyJ7StockId,
                ReservedAmount = 3
            };

            // reservation test
            stockService.ReserveStock(samsungGalaxyJ7Reservation);
            var res1 = stockService.GetCurrentlyAvailableUnits(galaxyJ7StockId);
            Console.WriteLine(res1 == 3 ? "StockService - Test01 - OK" : "StockService - Test01 - FAIL");

            // release test
            stockService.ReleaseStockReservations(martinCustomerId, galaxyJ7StockId);
            var res2 = stockService.GetCurrentlyAvailableUnits(galaxyJ7StockId);
            Console.WriteLine(res2 == 6 ? "StockService - Test02 - OK" : "StockService - Test02 - FAIL");

            // dispatch test
            stockService.ReserveStock(samsungGalaxyJ7Reservation);
            stockService.DispatchStock(martinCustomerId);
            var res3 = stockService.GetCurrentlyAvailableUnits(galaxyJ7StockId);
            Console.WriteLine(res3 == 3 ? "StockService - Test03 - OK" : "StockService - Test03 - FAIL");

            // order from distributor test
            stockService.OrderStockFromDistributor(galaxyJ7StockId, 3);
            var res4 = stockService.GetCurrentlyAvailableUnits(galaxyJ7StockId);
            Console.WriteLine(res4 == 6 ? "StockService - Test04 - OK" : "StockService - Test04 - FAIL");
        }

        private static void TestCheckoutService()
        {
            checkoutService = Container.Resolve<ICheckoutService>();
            salesService = Container.Resolve<ISalesService>();
            
            var customer = customerService.GetCustomer(1);

            var myOrder = new OrderDTO
            {
                CustomerID = customer.ID,
                Issued = DateTime.Now
            };

            var myItem1 = new OrderItemDTO
            {
                Product = productService.ListProducts(new ProductFilter
                {
                    ProductName = "Samsung"
                }).ResultsPage.First(),
                Quantity = 2
            };

            var myItem2 = new OrderItemDTO
            {                
                Product = productService.ListProducts(new ProductFilter
                {
                    ProductName = "Lumia 550"
                }).ResultsPage.First(),
                Quantity = 1
            };

            var shoppingCart = new List<OrderItemDTO> {myItem1, myItem2};
            
            checkoutService.ConfirmOrder(new OrderCreateDTO
            {
                CustomerId = customer.ID,
                Order = myOrder,
                OrderItems = shoppingCart
            }, () => Console.WriteLine("Checkout method ConfirmOrder worked"));
        
            var totalPrice = checkoutService.GetItemsTotalPrice(shoppingCart.ToArray());
            Console.WriteLine(totalPrice == 17840.00m ? "CheckoutService - Test01 - OK" : "CheckoutService - Test01 - FAIL");

            var calculatedTotalPrice = checkoutService.CalculateTotalPrice(shoppingCart);
            Console.WriteLine(calculatedTotalPrice == 17840.00m ? "CheckoutService - Test02 - OK" : "CheckoutService - Test02 - FAIL");

            var orders = customerService.ListOrders(new OrderFilter { CustomerId = 1 });
            var orderItems = salesService.ListItemsFromOrder(
                new OrderItemFilter
                {
                    OrderId = orders.First().ID
                });
        }

        private static void TestRatingService()
        {
            ratingService = Container.Resolve<IRatingService>();

            var reviewsBefore = ratingService.GetAllReviewsForProduct(1);

            ratingService.AddReview(new ProductReviewCreateDTO
            {
                ProductID = 1,
                Author = "Martin",
                UserRating = 8.5,
                Cons = "Small",
                Pros = "Nice design",
                Review = "It could be better, but it's OK"
            });
            ratingService.AddReview(new ProductReviewCreateDTO
            {
                ProductID = 1,
                Author = "Peter",
                UserRating = 6.5,
                Cons = "Big, horrible design, slow",
                Pros = "I gave it to my sister as birthday present",
                Review = "I don't like it"
            });

            var reviewsAfter = ratingService.GetAllReviewsForProduct(1);
            Console.WriteLine(reviewsAfter.Count() - reviewsBefore.Count() == 2 ? "RatingService - Test01 - OK" : "RatingService - Test01 - FAIL");
        }

        private static void TestPromoteService()
        {
            promoteService = Container.Resolve<IPromoteService>();
            customerService = Container.Resolve<ICustomerService>();

            var customer = customerService.GetCustomer(1);
            
            promoteService.CreateCoupon(customer.ID);
            promoteService.SendBirthDayCouponIfPossible(customer.ID);           
        }
    }
}
