﻿using System.Collections.Generic;
using System.Linq;
using System.Web.Mvc;
using BL.DTOs.Categories;
using BL.DTOs.Filters;
using BL.DTOs.Products;
using BL.DTOs.Reviews;
using BL.Facades;
using PL.Models;
using X.PagedList;

namespace PL.Controllers
{
    public class ProductsController : Controller
    {
        #region SessionKey constants

        private readonly string filterSessionKey = "filter";

        private readonly string categoryTreesSessionKey = "categoryTrees";

        #endregion

        #region Facades

        public ProductFacade ProductFacade { get; set; }

        #endregion

        #region ProductListActionMethods

        public ActionResult Index(int page = 1)
        {
            var filter = Session[filterSessionKey] as ProductFilter ?? new ProductFilter();
            var categoryTrees = Session[categoryTreesSessionKey] as IList<CategoryDTO>;

            var result = ProductFacade.GetProducts(filter, page);

            var model = InitializeProductListViewModel(result, categoryTrees);

            return View("ProductListView", model);
        }

        [HttpPost]
        public ActionResult Index(ProductListViewModel model)
        {
            model.Filter.CategoryIds = ProcessCategoryIds(model);

            Session[filterSessionKey] = model.Filter;
            Session[categoryTreesSessionKey] = model.Categories;

            var result = ProductFacade.GetProducts(model.Filter);

            var newModel = InitializeProductListViewModel(result, model.Categories);

            return View("ProductListView", newModel);
        }

        public ActionResult ClearFilter()
        {
            Session[filterSessionKey] = null;
            Session[categoryTreesSessionKey] = null;
            return RedirectToAction(nameof(Index));
        } 

        #endregion

        #region ProductDetailActionMethods

        public ActionResult Details(int id)
        {
            var model = ProductFacade.GetProductDetails(id);
            return View("ProductDetailView", model);
        }

        public ActionResult AddReview(int id)
        {
            return View("ProductReviewCreate", new ProductReviewCreateDTO { ProductID = id });
        }

        [HttpPost]
        public ActionResult AddReview(ProductReviewCreateDTO model)
        {
            ProductFacade.AddReview(model);
            return RedirectToAction(nameof(Details), new { id = model.ProductID });
        } 

        #endregion

        #region Helper methods

        /// <summary>
        /// Initializes new ProductListViewModel according to its parameters
        /// </summary>
        /// <param name="result">Product list query result containing products page and related data</param>
        /// <param name="categories">List of category trees</param>
        /// <returns>Initialized instance of ProductListViewModel</returns>
        private ProductListViewModel InitializeProductListViewModel(ProductListQueryResultDTO result, IList<CategoryDTO> categories = null)
        {
            return new ProductListViewModel
            {
                Products = new StaticPagedList<ProductDTO>(result.ResultsPage, result.RequestedPage, ProductFacade.ProductPageSize, result.TotalResultCount),
                Categories = categories ?? ProductFacade.GetAllCategories() as IList<CategoryDTO>,
                Filter = result.Filter
            };
        }

        /// <summary>
        /// Processes category IDs by filtering out unchecked categories
        /// </summary>
        /// <param name="model">model containing category trees</param>
        /// <returns>selected categories</returns>
        private static int[] ProcessCategoryIds(ProductListViewModel model)
        {
            var selectedCategoryIds = new List<int>();
            foreach (var categoryTreeRoot in model.Categories)
            {
                if (categoryTreeRoot.IsActive)
                {
                    selectedCategoryIds.Add(categoryTreeRoot.ID);
                    selectedCategoryIds.AddRange(model.Categories
                        .Where(node => node.ParentID == categoryTreeRoot.ID && node.IsActive)
                        .Select(node => node.ID));
                }                
            }
            var categoryIds = selectedCategoryIds.ToArray();
            return categoryIds;
        } 
        
        #endregion
    }
}
